luajalayout        = { }

luajalayout.module = {
   name           = "luajalayout",
   version        = 0.20,
   date           = "2010/03/16",
   description    = "Japanese text layout for LuaLaTeX.",
   author         = "Kazuki Maeda",
   copyright      = "Kazuki Maeda",
   license        = "MIT License"
}

local error, warning, info, log = luatexbase.provides_module(luajalayout.module)

utf = unicode.utf8

-- load auxiliary code (definition of some tables, etc.)
require("luajalayout_aux.lua")

luajalayout.process_input_buffer_start = false

function luajalayout.get_ja_range()
   local tmp = ""
   for k, v in pairs(luajalayout.lang_char_table[luajalayout.lang.ja]) do
      tmp = tmp .. "," .. tostring(v[1]) .. "-" .. tostring(v[2])
   end
   tex.print(utf.sub(tmp, 2))
end

-- language of a given character
local function get_lang(c)
   -- Japanese
   for k, v in pairs(luajalayout.lang_char_table[luajalayout.lang.ja]) do
      if v[1] <= c and c <= v[2] then
         return luajalayout.lang.ja
      end
   end
   
   -- other
   return luajalayout.lang.ot
end

local function get_char_table_values(char)
   if not luajalayout.char_table[char] then
      return {}
   else
      return luajalayout.char_table[char]
   end
end

-- TeX's \hss
local function get_hss()
   local hss = node.new(node.id("glue"))
   local hss_spec = node.new(node.id("glue_spec"))
   hss_spec.width = 0
   hss_spec.stretch = 65536
   hss_spec.stretch_order = 2
   hss_spec.shrink = 65536
   hss_spec.shrink_order = 2
   hss.spec = hss_spec
   return hss
end

local function get_glue(spec, em)
   if not em then em = 0 end
   local glue = node.new(node.id("glue"))
   local glue_spec = node.new(node.id("glue_spec"))
   if spec.width then glue_spec.width = spec.width*em; end
   if spec.stretch then glue_spec.stretch = spec.stretch*em; end
   if spec.shrink then glue_spec.shrink = spec.shrink*em; end
   glue.spec = glue_spec
   return glue
end

local function get_kern(spec, em)
   if not em then em = 0 end
   local kern = node.new(node.id("kern"))
   if spec.kern then kern.kern = spec.kern*em; end
   return kern
end


-- remove glues inserted after a Japanese character at the end of a line
local function remove_ja_eol(buffer)
   if luajalayout.process_input_buffer_start and
      tex.attribute.l_luajalayout_if_remove_ja_eol_attr > 0 and
      utf.len(buffer) > 0 and
      get_lang(utf.byte(buffer, utf.len(buffer))) == luajalayout.lang.ja then
      buffer = buffer .. "%"
   end
   return buffer
end

local function insert_ja_ja_glues(head, pn, cn)
   local pnsize, cnsize
   if not fonts.ids[pn.font].size then pnsize = 0 else pnsize = fonts.ids[pn.font].size end
   if not fonts.ids[cn.font].size then cnsize = 0 else cnsize = fonts.ids[cn.font].size end

   local c_pc = get_char_table_values(pn.char).ja_class
   local c_cc = get_char_table_values(cn.char).ja_class

   if not c_pc then c_pc = luajalayout.ja_class.other; end
   if not c_cc then c_cc = luajalayout.ja_class.other; end
   
   if luajalayout.ja_ja_glue_table[c_pc][c_cc] then
      if luajalayout.ja_ja_glue_table[c_pc][c_cc].kern then
         return node.insert_before(head, cn, get_kern(luajalayout.ja_ja_glue_table[c_pc][c_cc], cnsize))
      else         
         return node.insert_before(head, cn, get_glue(luajalayout.ja_ja_glue_table[c_pc][c_cc], cnsize))
      end
   else
      return node.insert_before(head, cn, get_glue(luajalayout.ja_ja_glue_other, (pnsize+cnsize)/2))      
   end
end

local function set_ja_width(head, cn, v_cn)
   local glyph = cn
   head, cn = node.remove(head, cn)
   glyph.next = nil
   
   local hbox_head = nil
   if v_cn.ja_width == luajalayout.ja_width.halfright 
      or v_cn.ja_width == luajalayout.ja_width.halfcenter then
      local hss_left = get_hss()
      hbox_head = hss_left
      hbox_head = node.insert_after(hbox_head, hss_left, glyph)
   else
      hbox_head = glyph
   end
   if v_cn.ja_width == luajalayout.ja_width.halfleft
      or v_cn.ja_width == luajalayout.ja_width.halfcenter then
      hbox_head = node.insert_after(hbox_head, glyph, get_hss())
   end
   
   local hbox
   if node.has_attribute(glyph, luatexbase.attributes.l_luajalayout_hang_punct_attr, 1) and 
      v_cn.ja_hang_punct then
      hbox = node.hpack(hbox_head, 0, "exactly")
   else
      hbox = node.hpack(hbox_head, 0.5*fonts.ids[glyph.font].size, "exactly") 
   end

   return node.insert_before(head, cn, hbox)
end

-- main process
local function node_filter(head, ...)
   local pn = nil
   local cn = head
   local mathflag = false

   while true do
      while cn and node.type(cn.id) == "whatsit" do
         cn = cn.next
      end

      if not cn then break; end

      local pn_next = cn
      local v_cn, v_pn

      if pn and node.type(pn.id) == "glyph" then
         v_pn = get_char_table_values(pn.char) 
      end
      if cn and node.type(cn.id) == "glyph" then 
         v_cn = get_char_table_values(cn.char)
      end

      if mathflag then
         if node.type(cn.id) == "math" then 
            mathflag = false
         end
      else
         if node.type(cn.id) == "math" then
            mathflag = true
         end

         if pn then
            -- hanging punctuation (insert glue)
            if node.has_attribute(cn, luatexbase.attributes.l_luajalayout_hang_punct_attr, 1) and 
               v_pn and v_pn.ja_hang_punct then
               local spec = {width = 0.5}
               node.insert_before(head, cn, get_glue(spec, fonts.ids[pn.font].size))
            end
            
            -- glyph and math
            if node.has_attribute(cn, luatexbase.attributes.l_luajalayout_if_insert_ot_ja_glue_attr, 1) and 
               node.type(pn.id) == "math" and v_cn and
               get_lang(cn.char) == luajalayout.lang.ja and
               v_cn.inhibit_ot_ja_glue ~= luajalayout.inhibit_ot_ja_glue.both and
               v_cn.inhibit_ot_ja_glue ~= luajalayout.inhibit_ot_ja_glue.before then
               node.insert_before(head, cn, get_glue(luajalayout.ot_ja_glue, fonts.ids[cn.font].size))   
            elseif node.has_attribute(cn, luatexbase.attributes.l_luajalayout_if_insert_ot_ja_glue_attr, 1) and 
                   node.type(cn.id) == "math" and v_pn and
                   get_lang(pn.char) == luajalayout.lang.ja and
                   v_pn.inhibit_ot_ja_glue ~= luajalayout.inhibit_ot_ja_glue.both and
                   v_pn.inhibit_ot_ja_glue ~= luajalayout.inhibit_ot_ja_glue.after then
               node.insert_before(head, cn, get_glue(luajalayout.ot_ja_glue, fonts.ids[pn.font].size))
            end
         
            -- glyph and glyph
            if v_pn and v_cn 
               and not (get_lang(pn.char) == luajalayout.lang.ot
                     and get_lang(cn.char) == luajalayout.lang.ot) then
               -- prebreakpenalty
               if v_cn.prebreakpenalty then
                  local penalty = node.new(node.id("penalty"))
                  penalty.penalty = v_cn.prebreakpenalty
                  node.insert_before(head, cn, penalty)
               end

               -- postbreakpenalty
               if v_pn.postbreakpenalty then
                  local penalty = node.new(node.id("penalty"))
                  penalty.penalty = v_pn.postbreakpenalty
                  node.insert_before(head, cn, penalty)
               end
              
               -- insert glue
               if node.has_attribute(cn, luatexbase.attributes.l_luajalayout_if_insert_ot_ja_glue_attr, 1) and
                  get_lang(pn.char) == luajalayout.lang.ot and
                  get_lang(cn.char) == luajalayout.lang.ja and
                  v_pn.inhibit_ot_ja_glue ~= luajalayout.inhibit_ot_ja_glue.after and
                  v_cn.inhibit_ot_ja_glue ~= luajalayout.inhibit_ot_ja_glue.both and
                  v_cn.inhibit_ot_ja_glue ~= luajalayout.inhibit_ot_ja_glue.before then
                  node.insert_before(head, cn, get_glue(luajalayout.ot_ja_glue, fonts.ids[cn.font].size))
               elseif node.has_attribute(cn, luatexbase.attributes.l_luajalayout_if_insert_ot_ja_glue_attr, 1) and
                      get_lang(pn.char) == luajalayout.lang.ja and
                      get_lang(cn.char) == luajalayout.lang.ot and
                      v_pn.inhibit_ot_ja_glue ~= luajalayout.inhibit_ot_ja_glue.both and
                      v_pn.inhibit_ot_ja_glue ~= luajalayout.inhibit_ot_ja_glue.after and
                      v_cn.inhibit_ot_ja_glue ~= luajalayout.inhibit_ot_ja_glue.before then
                  node.insert_before(head, cn, get_glue(luajalayout.ot_ja_glue, fonts.ids[pn.font].size))
               elseif node.has_attribute(cn, luatexbase.attributes.l_luajalayout_if_insert_ja_ja_glue_attr, 1) and
                      get_lang(pn.char) == luajalayout.lang.ja and
                      get_lang(cn.char) == luajalayout.lang.ja then
                      insert_ja_ja_glues(head, pn, cn)
               end               
            end
         end
        
         -- ja_width
         if node.has_attribute(cn, luatexbase.attributes.l_luajalayout_if_insert_ja_ja_glue_attr, 1) and 
            v_cn and v_cn.ja_width then
            head, cn = set_ja_width(head, cn, v_cn)
         end
      end
      pn = pn_next
      cn = cn.next
   end
   return head
end

luatexbase.add_to_callback("pre_linebreak_filter", 
                           node_filter,
                           "luajalayout.pre_linebreak_filter",
                           luatexbase.priority_in_callback(
                              'pre_linebreak_filter',
                              'luaotfload.pre_linebreak_filter'))
luatexbase.add_to_callback("hpack_filter", 
                           node_filter,
                           "luajalayout.hpack_filter",
                           luatexbase.priority_in_callback(
                              'hpack_filter',
                              'luaotfload.hpack_filter'))
luatexbase.add_to_callback("process_input_buffer",
                           remove_ja_eol,
                           "luajalayout.process_input_buffer")
